#!/usr/bin/env python2
# -*- coding: utf8 -*-

__author__ = "Stefan Huber"
__copyright__ = "Copyright 2015"

__version__ = "2.0"
__license__ = "LGPL3"


import sys, getopt, os

import numpy as np
import matplotlib as mpl
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker


class Model:

    def __init__(self, itmodel=None):
        """itmodel is the income tax model. If None, it is the current model
        from 2014. The current model would be (almost) equivalent [(60000,
        0.5), (25000, 0.4314286), (11000, 0.265)]"""

        self.itmodel = itmodel
        if self.itmodel is not None:
            self.itmodel = list(sorted(self.itmodel, reverse=True))

        print(self.itmodel)

    def uiPerMonth(self, gross):
        """Returns unemployment insurance rates (in percent) from a specific gross
        income per month."""
        if gross <= 1219:
            return 0.0
        elif gross <= 1330:
            return 1.0
        elif gross <= 1497:
            return 2.0
        else:
            return 3.0

    def siPerMonth(self, month, gross):
        """Returns social insurance dues from a specific gross income per month for
        the given month. Month is between 1 and 14."""

        if gross <= 386.80:
            return 0

        # retirement pension insurance, health insurance, unemployment insurace
        rates = 10.25 + 3.82 + self.uiPerMonth(gross)
        if month <= 12:
            # chamber of labour dues, residential building support
            rates += 0.5 + 0.5

        return rates/100.0 * min(gross, 4400)

    def incometaxPerYear(self, yearref):
        """Income tax of the reference (gross per year minus reductions)."""
        assert(self.itmodel is not None)

        for (bound, rate) in self.itmodel:
            if yearref > bound:
                it = (yearref-bound)*rate + self.incometaxPerYear(bound)
                return it
        return 0

    def itPerMonthFromModel(self, monthgross):
        assert(self.itmodel is not None)

        # Do not know why 12 instead of 14. But it matches better.
        yeargross = 12*monthgross
        yearsi = self.siPerYear(monthgross)
        yearref = yeargross - yearsi

        it = self.incometaxPerYear(yearref)
        return it/12

    def itertable(self, ref):
        """Returns the income tax of the reference (gross minus reductions)
        according to the effective rates table (Effektiv-Tarif-Tabelle)."""

        net = None

        if ref <= 1011.43:
            net = 0.0
        elif ref <= 2099.33:
            net = ref * 0.365 - 369.173
        elif ref <= 5016:
            net = ref * 0.4321429 - 510.129
        else:
            net = ref * 0.5 - 850.5

        return net

    def itPerMonth(self, month, gross):
        """Returns income tax from a specific gross income per month for the given
        month. Month is between 1 and 14."""

        # The reference value for the income tax
        ref = gross - self.siPerMonth(month, gross)

        if month > 12:
            # Reduction of taxes for special renumeration
            if month == 13:
                ref -= 620

            # The number 1050 was reverse engineered from "Brutto-Netto-Rechner"
            # from the webpage of the ministry of finance.
            if gross <= 1050:
                return 0
            else:
                return ref * 6.0/100.0
        else:
            if self.itmodel is None:
                return self.itertable(ref)
            else:
                return self.itPerMonthFromModel(gross)

    def netPerMonth(self, month, gross):
        """Returns net income from a specific gross income per month for the given
        month. Month is between 1 and 14."""
        return gross - self.siPerMonth(month, gross) - self.itPerMonth(month, gross)

    def netPerYear(self, gross):
        """Returns net income per year of given per-month gross income."""
        return sum([self.netPerMonth(m, gross) for m in range(1, 15)])

    def itPerYear(self, gross):
        """Returns income tax per year of given per-month gross income."""
        return sum([self.itPerMonth(m, gross) for m in range(1, 15)])

    def siPerYear(self, gross):
        """Returns socual insurance tax per year of given per-month gross income."""
        return sum([self.siPerMonth(m, gross) for m in range(1, 15)])


def printTaxesTable(gross):
    """Print a table with taxes for given per-month gross income."""

    m = Model()

    print("          |  1.-12. month |  13. month |  14. month ||       year")
    print("          |---------------+------------+------------++------------")
    print("gross     | %13.2f | %10.2f | %10.2f || %10.2f" % \
            (gross, gross, gross, 14*gross))
    print("          |---------------+------------+------------++------------")
    print("soc. ins. | %13.2f | %10.2f | %10.2f || %10.2f" % \
            (m.siPerMonth(1, gross),
                m.siPerMonth(13, gross),
                m.siPerMonth(14, gross),
                m.siPerYear(gross)))
    print("inc. tax  | %13.2f | %10.2f | %10.2f || %10.2f" % \
            (m.itPerMonth(1, gross),
                m.itPerMonth(13, gross),
                m.itPerMonth(14, gross),
                m.itPerYear(gross)))
    print("          |---------------+------------+------------++------------")
    print("net       | %13.2f | %10.2f | %10.2f || %10.2f" % \
            (m.netPerMonth(1, gross),
                m.netPerMonth(13, gross),
                m.netPerMonth(14, gross),
                m.netPerYear(gross)))

def diffQuotient(func, delta=1.0):
    return lambda x: (func(x+delta) - func(x))/delta


def showPlots():

    print("Showing plots...")

    m = Model()

    plt.figure(1, figsize=(6.5, 8), dpi=80, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)
    t = np.arange(15, 7000, 10)

    sp = plt.subplot(311)
    plt.ylabel(u'Jahresbeträge')
    net = map(m.netPerYear, t)
    si = map(m.siPerYear, t)
    it = map(m.itPerYear, t)
    plt.plot(t, net, label='Nettoeinkommen', linewidth=2.0)
    plt.plot(t, si, label='Sozialversicherung', linewidth=2.0, color="lightgreen")
    plt.plot(t, it, label='Lohnsteuer', linewidth=2.0, color="green")
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(312)
    plt.ylabel(u'Abl. der Jahresbeträge')
    net = map(diffQuotient(m.netPerYear), t)
    si = map(diffQuotient(m.siPerYear), t)
    it = map(diffQuotient(m.itPerYear), t)
    plt.plot(t, net, label='Nettoeinkommen', linewidth=2.0)
    plt.plot(t, si, label='Sozialversicherung', linewidth=2.0, color="lightgreen")
    plt.plot(t, it, label='Lohnsteuer', linewidth=2.0, color="green")
    sp.set_ylim(0, 15)
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(313)
    plt.ylabel(u'Relative Jahresbeträge')
    net = map(lambda x: m.netPerYear(x)/x/14, t)
    si = map(lambda x: m.siPerYear(x)/x/14, t)
    it = map(lambda x: m.itPerYear(x)/x/14, t)
    plt.plot(t, net, label='Nettoeinkommen', linewidth=2.0)
    plt.plot(t, si, label='Sozialversicherung', linewidth=2.0, color="lightgreen")
    plt.plot(t, it, label='Lohnsteuer', linewidth=2.0, color="green")
    sp.set_ylim(0, 1.1)
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")


    plt.figure(2, figsize=(6.5, 8), dpi=80, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)

    sp = plt.subplot(311)
    plt.ylabel(u'Monatsbeträge')
    net = map(lambda x: m.netPerMonth(1, x), t)
    si = map(lambda x: m.siPerMonth(1, x), t)
    it = map(lambda x: m.itPerMonth(1, x), t)
    plt.plot(t, net, label='Nettoeinkommen', linewidth=2.0)
    plt.plot(t, si, label='Sozialversicherung', linewidth=2.0, color="lightgreen")
    plt.plot(t, it, label='Lohnsteuer', linewidth=2.0, color="green")
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(312)
    plt.ylabel(u'Abl. der Monatsbeträge')
    net = map(diffQuotient(lambda x: m.netPerMonth(1,x)), t)
    si = map(diffQuotient(lambda x: m.siPerMonth(1,x)), t)
    it = map(diffQuotient(lambda x: m.itPerMonth(1,x)), t)
    plt.plot(t, net, label='Nettoeinkommen', linewidth=2.0)
    plt.plot(t, si, label='Sozialversicherung', linewidth=2.0, color="lightgreen")
    plt.plot(t, it, label='Lohnsteuer', linewidth=2.0, color="green")
    sp.set_ylim(0, 1.1)
    sp.grid(True)
    plt.legend(loc=1, prop={'size': 11})
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(313)
    plt.ylabel(u'Relative Monatsbeträge')
    net = map(lambda x: m.netPerMonth(1, x)/x, t)
    si = map(lambda x: m.siPerMonth(1, x)/x, t)
    it = map(lambda x: m.itPerMonth(1, x)/x, t)
    plt.plot(t, net, label='Nettoeinkommen', linewidth=2.0)
    plt.plot(t, si, label='Sozialversicherung', linewidth=2.0, color="lightgreen")
    plt.plot(t, it, label='Lohnsteuer', linewidth=2.0, color="green")
    sp.set_ylim(0, 1.1)
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")


    plt.figure(3, figsize=(6.5, 3), dpi=80, facecolor='w')
    plt.subplots_adjust(left=0.15, right=0.9, bottom=0.15, top=0.95)

    sp = plt.subplot(111)
    plt.ylabel(u'Nettogehälter')
    t = np.arange(1000, 2000, 10)
    net = map(lambda x: m.netPerMonth(1, x), t)
    net13 = map(lambda x: m.netPerMonth(13, x), t)
    net14 = map(lambda x: m.netPerMonth(14, x), t)
    plt.plot(t, net, label='1.-12. Monat', linewidth=2.0)
    plt.plot(t, net13, label='13. Monat', linewidth=2.0, color="lightgreen")
    plt.plot(t, net14, label='14. Monat', linewidth=2.0, color="green")
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    plt.show()


def showPlotsCompareModels():

    def to_percent(y, pos):
        s = str(100 * y)
        if mpl.rcParams['text.usetex']:
            return s + r'$\%$'
        else:
            return s + '%'

    print("Comparing models...")

    accmodel = Model()
    currmodel = Model([(11000, 0.365), (25000, 0.4314), (60000, 0.5)])
    spoemodel = Model([(11000, 0.25), (20000, 0.34), (30000, 0.38), (45000,
        0.43), (60000, 0.47), (80000, 0.5)])
    oevpmodel = Model([(11000, 0.25), (16000, 0.35), (30000, 0.42), (60000,
        0.47), (100000, 0.5)])

    #modells = [(currmodel, "Aktuell", "blue"), (spoemodel, u"SPÖ", "lightgreen"),
            #(oevpmodel, u"ÖVP", "orange")]
    #basemodel = Model()
    basemodel = currmodel
    modells = [(basemodel, "Aktuell", "grey"), (spoemodel, u"SPÖ", "red"), (oevpmodel, u"ÖVP", "black")]

    t = np.arange(1000, 10000, 50)


    plt.figure(1, figsize=(6.5, 8), dpi=60, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)

    sp = plt.subplot(211)
    plt.ylabel(u'Jahresnettogehalt')
    for m, l, c in modells:
        f = map(lambda x: m.netPerYear(x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(212)
    plt.ylabel(u'Monatsnettogehalt')
    for m, l, c in modells:
        f = map(lambda x: m.netPerMonth(1, x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")


    plt.figure(2, figsize=(6.5, 8), dpi=60, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)

    sp = plt.subplot(211)
    plt.ylabel(u'Diff. Jahresnettogehalt')
    for m, l, c in modells[1:]:
        f = map(lambda x: m.netPerYear(x) - basemodel.netPerYear(x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    f = map(lambda x: spoemodel.netPerYear(x) - oevpmodel.netPerYear(x), t)
    plt.plot(t, f, label=u"SPÖ-ÖVP", linewidth=1.0, color='grey')
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(212)
    plt.ylabel(u'Diff. Monatsnettogehalt')
    for m, l, c in modells[1:]:
        f = map(lambda x: m.netPerMonth(1, x) - basemodel.netPerMonth(1, x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    f = map(lambda x: spoemodel.netPerMonth(1, x) - oevpmodel.netPerMonth(1, x), t)
    plt.plot(t, f, label=u"SPÖ-ÖVP", linewidth=1.0, color='grey')
    plt.legend(loc=2, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")


    plt.figure(3, figsize=(6.5, 8), dpi=60, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)

    sp = plt.subplot(211)
    plt.gca().yaxis.set_major_formatter(ticker.FuncFormatter(to_percent))
    plt.ylabel(u'Rel. Diff. Jahresnettogehalt')
    for m, l, c in modells[1:]:
        f = map(lambda x: (m.netPerYear(x) - basemodel.netPerYear(x))/basemodel.netPerYear(x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    f = map(lambda x: (spoemodel.netPerYear(x) - oevpmodel.netPerYear(x))/basemodel.netPerYear(x), t)
    plt.plot(t, f, label=u"SPÖ-ÖVP", linewidth=1.0, color='grey')
    sp.set_ylim(0, 0.065)
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(212)
    plt.gca().yaxis.set_major_formatter(ticker.FuncFormatter(to_percent))
    plt.ylabel(u'Rel. Diff. Monatsnettogehalt')
    for m, l, c in modells[1:]:
        f = map(lambda x: (m.netPerMonth(1, x) - basemodel.netPerMonth(1, x))/basemodel.netPerMonth(1, x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    f = map(lambda x: (spoemodel.netPerMonth(1, x) - oevpmodel.netPerMonth(1, x))/basemodel.netPerMonth(1, x), t)
    plt.plot(t, f, label=u"SPÖ-ÖVP", linewidth=1.0, color='grey')
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")


    plt.figure(4, figsize=(6.5, 8), dpi=60, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)

    sp = plt.subplot(211)
    plt.gca().yaxis.set_major_formatter(ticker.FuncFormatter(to_percent))
    plt.ylabel(u'Modellfehler Jahresnettogehalt')
    f = map(lambda x: (currmodel.netPerYear(x) - accmodel.netPerYear(x))/accmodel.netPerYear(x), t)
    plt.plot(t, f, linewidth=2.0, color='black')
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")

    sp = plt.subplot(212)
    plt.gca().yaxis.set_major_formatter(ticker.FuncFormatter(to_percent))
    plt.ylabel(u'Modellfehler Monatsnettogehalt')
    f = map(lambda x: (currmodel.netPerMonth(1, x) - accmodel.netPerMonth(1, x))/accmodel.netPerMonth(1, x), t)
    plt.plot(t, f, linewidth=2.0, color='black')
    plt.legend(loc=1, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Brutto-Monatsgehalt")


    t = np.arange(1000, 120000, 1000)

    plt.figure(5, figsize=(6.5, 8), dpi=60, facecolor='w')
    plt.subplots_adjust(left=0.13, right=0.95, bottom=0.06, top=0.98)

    sp = plt.subplot(211)
    plt.gca().yaxis.set_major_formatter(ticker.FuncFormatter(to_percent))
    plt.ylabel(u'Einkommenssteuersatz')
    for m, l, c in modells:
        f = map(diffQuotient(lambda x: m.incometaxPerYear(x)), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    sp.set_ylim(0, 0.55)
    plt.legend(loc=0, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Bemessungsgrundlage")

    sp = plt.subplot(212)
    plt.ylabel(u'Lohnsteuer')
    for m, l, c in modells:
        f = map(lambda x: m.incometaxPerYear(x), t)
        plt.plot(t, f, label=l, linewidth=2.0, color=c)
    plt.legend(loc=0, prop={'size': 11})
    plt.grid(True)
    plt.xlabel("Bemessungsgrundlage")

    plt.show()


def printUsage():
    print("""Usage:
  {0} -t gross-per-month
  {0} -p
  {0} -m
  {0} -h""".format(sys.argv[0]))


if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hpmt:")

        for opt, arg in opts:
            if opt == "-h":
                printUsage()
                sys.exit(os.EX_OK)
            if opt == "-p":
                showPlots()
            if opt == "-m":
                showPlotsCompareModels()
            if opt == "-t":
                printTaxesTable(int(arg))

    except getopt.GetoptError as e:
        print("Error parsing arguments:", e)
        printUsage()

